@@ -55,7 +55,7 @@ static int mnt_context_append_additional_mount(struct libmnt_context *cxt,
 	assert(cxt);
 	assert(ad);
 
-	DBG(CXT, mnt_debug_h(cxt,
+	DBG(CXT, ul_debugobj(cxt,
 			"mount: add additional flag: 0x%08lx",
 			ad->mountflags));
 
@@ -73,7 +73,7 @@ static int init_propagation(struct libmnt_context *cxt)
 	if (!opts)
 		return 0;
 
-	DBG(CXT, mnt_debug_h(cxt, "mount: initialize additional propagation mounts"));
+	DBG(CXT, ul_debugobj(cxt, "mount: initialize additional propagation mounts"));
 
 	maps[0] = mnt_get_builtin_optmap(MNT_LINUX_MAP);
 
@@ -102,6 +102,30 @@ static int init_propagation(struct libmnt_context *cxt)
 	return 0;
 }
 
+#if defined(HAVE_LIBSELINUX) || defined(HAVE_SMACK)
+struct libmnt_optname {
+	const char *name;
+	size_t namesz;
+};
+
+#define DEF_OPTNAME(n)		{ .name = n, .namesz = sizeof(n) - 1 }
+#define DEF_OPTNAME_LAST	{ .name = NULL }
+
+static int is_option(const char *name, size_t namesz,
+		     const struct libmnt_optname *names)
+{
+	const struct libmnt_optname *p;
+
+	for (p = names; p && p->name; p++) {
+		if (p->namesz == namesz
+		    && strncmp(name, p->name, namesz) == 0)
+			return 1;
+	}
+
+	return 0;
+}
+#endif /* HAVE_LIBSELINUX || HAVE_SMACK */
+
 /*
  * this has to be called after mnt_context_evaluate_permissions()
  */
@@ -114,6 +138,25 @@ static int fix_optstr(struct libmnt_context *cxt)
 	struct libmnt_fs *fs;
 #ifdef HAVE_LIBSELINUX
 	int se_fix = 0, se_rem = 0;
+	static const struct libmnt_optname selinux_options[] = {
+		DEF_OPTNAME("context"),
+		DEF_OPTNAME("fscontext"),
+		DEF_OPTNAME("defcontext"),
+		DEF_OPTNAME("rootcontext"),
+		DEF_OPTNAME("seclabel"),
+		DEF_OPTNAME_LAST
+	};
+#endif
+#ifdef HAVE_SMACK
+	int sm_rem = 0;
+	static const struct libmnt_optname smack_options[] = {
+		DEF_OPTNAME("smackfsdef"),
+		DEF_OPTNAME("smackfsfloor"),
+		DEF_OPTNAME("smackfshat"),
+		DEF_OPTNAME("smackfsroot"),
+		DEF_OPTNAME("smackfstransmute"),
+		DEF_OPTNAME_LAST
+	};
 #endif
 	assert(cxt);
 	assert(cxt->fs);
@@ -124,7 +167,7 @@ static int fix_optstr(struct libmnt_context *cxt)
 	if (!cxt->fs || (cxt->flags & MNT_FL_MOUNTOPTS_FIXED))
 		return 0;
 
-	DBG(CXT, mnt_debug_h(cxt, "mount: fixing optstr"));
+	DBG(CXT, ul_debugobj(cxt, "mount: fixing optstr"));
 
 	fs = cxt->fs;
 
@@ -149,13 +192,13 @@ static int fix_optstr(struct libmnt_context *cxt)
 	/*
 	 * Sync mount options with mount flags
 	 */
-	DBG(CXT, mnt_debug_h(cxt, "mount: fixing vfs optstr"));
+	DBG(CXT, ul_debugobj(cxt, "mount: fixing vfs optstr"));
 	rc = mnt_optstr_apply_flags(&fs->vfs_optstr, cxt->mountflags,
 				mnt_get_builtin_optmap(MNT_LINUX_MAP));
 	if (rc)
 		goto done;
 
-	DBG(CXT, mnt_debug_h(cxt, "mount: fixing user optstr"));
+	DBG(CXT, ul_debugobj(cxt, "mount: fixing user optstr"));
 	rc = mnt_optstr_apply_flags(&fs->user_optstr, cxt->user_mountflags,
 				mnt_get_builtin_optmap(MNT_USERSPACE_MAP));
 	if (rc)
@@ -169,7 +212,6 @@ static int fix_optstr(struct libmnt_context *cxt)
 		free(fs->user_optstr);
 		fs->user_optstr = NULL;
 	}
-
 	if (cxt->mountflags & MS_PROPAGATION) {
 		rc = init_propagation(cxt);
 		if (rc)
@@ -199,12 +241,14 @@ static int fix_optstr(struct libmnt_context *cxt)
 
 	if (!se_rem) {
 		/* de-duplicate SELinux options */
-		mnt_optstr_deduplicate_option(&fs->fs_optstr, "context");
-		mnt_optstr_deduplicate_option(&fs->fs_optstr, "fscontext");
-		mnt_optstr_deduplicate_option(&fs->fs_optstr, "defcontext");
-		mnt_optstr_deduplicate_option(&fs->fs_optstr, "rootcontext");
-		mnt_optstr_deduplicate_option(&fs->fs_optstr, "seclabel");
+		const struct libmnt_optname *p;
+		for (p = selinux_options; p && p->name; p++)
+			mnt_optstr_deduplicate_option(&fs->fs_optstr, p->name);
 	}
+#endif
+#ifdef HAVE_SMACK
+	if (access("/sys/fs/smackfs", F_OK) != 0)
+		sm_rem = 1;
 #endif
 	while (!mnt_optstr_next_option(&next, &name, &namesz, &val, &valsz)) {
 
@@ -213,12 +257,9 @@ static int fix_optstr(struct libmnt_context *cxt)
 		else if (namesz == 3 && !strncmp(name, "gid", 3))
 			rc = mnt_optstr_fix_gid(&fs->fs_optstr, val, valsz, &next);
 #ifdef HAVE_LIBSELINUX
-		else if ((se_rem || se_fix) &&
-			 namesz >= 7 && (!strncmp(name, "context", 7) ||
-					 !strncmp(name, "fscontext", 9) ||
-					 !strncmp(name, "defcontext", 10) ||
-					 !strncmp(name, "rootcontext", 11) ||
-					 !strncmp(name, "seclabel", 8))) {
+		else if ((se_rem || se_fix)
+			 && is_option(name, namesz, selinux_options)) {
+
 			if (se_rem) {
 				/* remove context= option */
 				next = name;
@@ -231,6 +272,15 @@ static int fix_optstr(struct libmnt_context *cxt)
 				rc = mnt_optstr_fix_secontext(&fs->fs_optstr,
 							val, valsz, &next);
 		}
+#endif
+#ifdef HAVE_SMACK
+		else if (sm_rem && is_option(name, namesz, smack_options)) {
+
+			next = name;
+			rc = mnt_optstr_remove_option_at(&fs->fs_optstr,
+					name,
+					val ? val + valsz : name + namesz);
+		}
 #endif
 		if (rc)
 			goto done;
@@ -246,7 +296,7 @@ static int fix_optstr(struct libmnt_context *cxt)
 done:
 	cxt->flags |= MNT_FL_MOUNTOPTS_FIXED;
 
-	DBG(CXT, mnt_debug_h(cxt, "fixed options [rc=%d]: "
+	DBG(CXT, ul_debugobj(cxt, "fixed options [rc=%d]: "
 		"vfs: '%s' fs: '%s' user: '%s', optstr: '%s'", rc,
 		fs->vfs_optstr, fs->fs_optstr, fs->user_optstr, fs->optstr));
 
@@ -270,7 +320,7 @@ static int generate_helper_optstr(struct libmnt_context *cxt, char **optstr)
 	assert(cxt->fs);
 	assert(optstr);
 
-	DBG(CXT, mnt_debug_h(cxt, "mount: generate helper mount options"));
+	DBG(CXT, ul_debugobj(cxt, "mount: generate helper mount options"));
 
 	*optstr = mnt_fs_strdup_options(cxt->fs);
 	if (!*optstr)
@@ -348,7 +398,7 @@ static int evaluate_permissions(struct libmnt_context *cxt)
 	if (!cxt->fs)
 		return 0;
 
-	DBG(CXT, mnt_debug_h(cxt, "mount: evaluating permissions"));
+	DBG(CXT, ul_debugobj(cxt, "mount: evaluating permissions"));
 
 	mnt_context_get_user_mflags(cxt, &u_flags);
 
@@ -364,7 +414,7 @@ static int evaluate_permissions(struct libmnt_context *cxt)
 		 */
 		if (!mnt_context_tab_applied(cxt))
 		{
-			DBG(CXT, mnt_debug_h(cxt, "perms: fstab not applied, ignore user mount"));
+			DBG(CXT, ul_debugobj(cxt, "perms: fstab not applied, ignore user mount"));
 			return -EPERM;
 		}
 
@@ -381,7 +431,7 @@ static int evaluate_permissions(struct libmnt_context *cxt)
 			if (!mnt_optstr_get_option(cxt->fs->user_optstr,
 					"user", NULL, &valsz) && valsz) {
 
-				DBG(CXT, mnt_debug_h(cxt, "perms: user=<name> detected, ignore"));
+				DBG(CXT, ul_debugobj(cxt, "perms: user=<name> detected, ignore"));
 				cxt->user_mountflags &= ~MNT_MS_USER;
 			}
 		}
@@ -407,7 +457,7 @@ static int evaluate_permissions(struct libmnt_context *cxt)
 				srcpath = xsrc;
 			}
 			if (!srcpath) {
-				DBG(CXT, mnt_debug_h(cxt, "perms: src undefined"));
+				DBG(CXT, ul_debugobj(cxt, "perms: src undefined"));
 				return -EPERM;
 			}
 
@@ -423,7 +473,7 @@ static int evaluate_permissions(struct libmnt_context *cxt)
 		}
 
 		if (!(cxt->user_mountflags & (MNT_MS_USER | MNT_MS_USERS))) {
-			DBG(CXT, mnt_debug_h(cxt, "permissions evaluation ends with -EPERMS"));
+			DBG(CXT, ul_debugobj(cxt, "permissions evaluation ends with -EPERMS"));
 			return -EPERM;
 		}
 	}
@@ -492,7 +542,7 @@ static int exec_helper(struct libmnt_context *cxt)
 	assert(cxt->helper);
 	assert((cxt->flags & MNT_FL_MOUNTFLAGS_MERGED));
 
-	DBG(CXT, mnt_debug_h(cxt, "mount: executing helper %s", cxt->helper));
+	DBG(CXT, ul_debugobj(cxt, "mount: executing helper %s", cxt->helper));
 
 	rc = generate_helper_optstr(cxt, &o);
 	if (rc)
@@ -518,12 +568,7 @@ static int exec_helper(struct libmnt_context *cxt)
 		args[i++] = mnt_fs_get_srcpath(cxt->fs);/* 2 */
 		args[i++] = mnt_fs_get_target(cxt->fs);	/* 3 */
 
-		/*
-		 * TODO: remove the exception for "nfs", -s is documented
-		 *       for years and should be usable everywhere.
-		 */
-		if (mnt_context_is_sloppy(cxt) &&
-		    type && startswith(type, "nfs"))
+		if (mnt_context_is_sloppy(cxt))
 			args[i++] = "-s";		/* 4 */
 		if (mnt_context_is_fake(cxt))
 			args[i++] = "-f";		/* 5 */
@@ -540,11 +585,9 @@ static int exec_helper(struct libmnt_context *cxt)
 			args[i++] = type;		/* 11 */
 		}
 		args[i] = NULL;				/* 12 */
-#ifdef CONFIG_LIBMOUNT_DEBUG
 		for (i = 0; args[i]; i++)
-			DBG(CXT, mnt_debug_h(cxt, "argv[%d] = \"%s\"",
+			DBG(CXT, ul_debugobj(cxt, "argv[%d] = \"%s\"",
 							i, args[i]));
-#endif
 		DBG_FLUSH;
 		execv(cxt->helper, (char * const *) args);
 		exit(EXIT_FAILURE);
@@ -555,7 +598,7 @@ static int exec_helper(struct libmnt_context *cxt)
 		wait(&st);
 		cxt->helper_status = WIFEXITED(st) ? WEXITSTATUS(st) : -1;
 
-		DBG(CXT, mnt_debug_h(cxt, "%s executed [status=%d]",
+		DBG(CXT, ul_debugobj(cxt, "%s executed [status=%d]",
 					cxt->helper, cxt->helper_status));
 		cxt->helper_exec_status = rc = 0;
 		break;
@@ -563,7 +606,7 @@ static int exec_helper(struct libmnt_context *cxt)
 
 	case -1:
 		cxt->helper_exec_status = rc = -errno;
-		DBG(CXT, mnt_debug_h(cxt, "fork() failed"));
+		DBG(CXT, ul_debugobj(cxt, "fork() failed"));
 		break;
 	}
 
@@ -589,7 +632,7 @@ static int do_mount_additional(struct libmnt_context *cxt,
 		struct libmnt_addmount *ad =
 				list_entry(p, struct libmnt_addmount, mounts);
 
-		DBG(CXT, mnt_debug_h(cxt, "mount(2) changing flag: 0x%08lx %s",
+		DBG(CXT, ul_debugobj(cxt, "mount(2) changing flag: 0x%08lx %s",
 				ad->mountflags,
 				ad->mountflags & MS_REC ? " (recursive)" : ""));
 
@@ -598,7 +641,7 @@ static int do_mount_additional(struct libmnt_context *cxt,
 		if (rc) {
 			if (syserr)
 				*syserr = -errno;
-			DBG(CXT, mnt_debug_h(cxt,
+			DBG(CXT, ul_debugobj(cxt,
 					"mount(2) failed [errno=%d %m]",
 					errno));
 			return rc;
@@ -653,7 +696,7 @@ static int do_mount(struct libmnt_context *cxt, const char *try_type)
 	if (!src) {
 		/* unnecessary, should be already resolved in
 		 * mnt_context_prepare_srcpath(), but to be sure... */
-		DBG(CXT, mnt_debug_h(cxt, "WARNING: source is NULL -- using \"none\"!"));
+		DBG(CXT, ul_debugobj(cxt, "WARNING: source is NULL -- using \"none\"!"));
 		src = "none";
 	}
 	type = try_type ? : mnt_fs_get_fstype(cxt->fs);
@@ -663,7 +706,7 @@ static int do_mount(struct libmnt_context *cxt, const char *try_type)
 	if (try_type)
 		flags |= MS_SILENT;
 
-	DBG(CXT, mnt_debug_h(cxt, "%smount(2) "
+	DBG(CXT, ul_debugobj(cxt, "%smount(2) "
 			"[source=%s, target=%s, type=%s, "
 			" mountflags=0x%08lx, mountdata=%s]",
 			mnt_context_is_fake(cxt) ? "(FAKE) " : "",
@@ -688,11 +731,11 @@ static int do_mount(struct libmnt_context *cxt, const char *try_type)
 		 */
 		if (mount(src, target, type, flags, cxt->mountdata)) {
 			cxt->syscall_status = -errno;
-			DBG(CXT, mnt_debug_h(cxt, "mount(2) failed [errno=%d %m]",
+			DBG(CXT, ul_debugobj(cxt, "mount(2) failed [errno=%d %m]",
 							-cxt->syscall_status));
 			return -cxt->syscall_status;
 		}
-		DBG(CXT, mnt_debug_h(cxt, "mount(2) success"));
+		DBG(CXT, ul_debugobj(cxt, "mount(2) success"));
 		cxt->syscall_status = 0;
 
 		/*
@@ -730,18 +773,36 @@ static int do_mount_by_pattern(struct libmnt_context *cxt, const char *pattern)
 		 */
 		char *p, *p0;
 
-		DBG(CXT, mnt_debug_h(cxt, "trying to mount by FS pattern list"));
+		DBG(CXT, ul_debugobj(cxt, "trying to mount by FS pattern list '%s'", pattern));
 
 		p0 = p = strdup(pattern);
 		if (!p)
 			return -ENOMEM;
 		do {
+			char *autotype = NULL;
 			char *end = strchr(p, ',');
+
 			if (end)
 				*end = '\0';
-			rc = do_mount(cxt, p);
-			p = end ? end + 1 : NULL;
 
+			DBG(CXT, ul_debugobj(cxt, "-->trying '%s'", p));
+
+			/* Let's support things like "udf,iso9660,auto" */
+			if (strcmp(p, "auto") == 0) {
+				rc = mnt_context_guess_srcpath_fstype(cxt, &autotype);
+				if (rc) {
+					DBG(CXT, ul_debugobj(cxt, "failed to guess FS type"));
+					free(p0);
+					return rc;
+				}
+				p = autotype;
+				DBG(CXT, ul_debugobj(cxt, "   --> '%s'", p));
+			}
+
+			if (p)
+				rc = do_mount(cxt, p);
+			p = end ? end + 1 : NULL;
+			free(autotype);
 		} while (!mnt_context_get_status(cxt) && p);
 
 		free(p0);
@@ -753,7 +814,7 @@ static int do_mount_by_pattern(struct libmnt_context *cxt, const char *pattern)
 	/*
 	 * try /etc/filesystems and /proc/filesystems
 	 */
-	DBG(CXT, mnt_debug_h(cxt, "trying to mount by filesystems lists"));
+	DBG(CXT, ul_debugobj(cxt, "trying to mount by filesystems lists"));
 
 	rc = mnt_get_filesystems(&filesystems, neg ? pattern : NULL);
 	if (rc)
@@ -800,7 +861,7 @@ int mnt_context_prepare_mount(struct libmnt_context *cxt)
 
 	cxt->action = MNT_ACT_MOUNT;
 
-	DBG(CXT, mnt_debug_h(cxt, "mount: preparing"));
+	DBG(CXT, ul_debugobj(cxt, "mount: preparing"));
 
 	rc = mnt_context_apply_fstab(cxt);
 	if (!rc)
@@ -818,7 +879,7 @@ int mnt_context_prepare_mount(struct libmnt_context *cxt)
 	if (!rc)
 		rc = mnt_context_prepare_helper(cxt, "mount", NULL);
 	if (rc) {
-		DBG(CXT, mnt_debug_h(cxt, "mount: preparing failed"));
+		DBG(CXT, ul_debugobj(cxt, "mount: preparing failed"));
 		return rc;
 	}
 	cxt->flags |= MNT_FL_PREPARED;
@@ -860,7 +921,7 @@ int mnt_context_do_mount(struct libmnt_context *cxt)
 	assert((cxt->flags & MNT_FL_PREPARED));
 	assert((cxt->action == MNT_ACT_MOUNT));
 
-	DBG(CXT, mnt_debug_h(cxt, "mount: do mount"));
+	DBG(CXT, ul_debugobj(cxt, "mount: do mount"));
 
 	if (!(cxt->flags & MNT_FL_MOUNTDATA))
 		cxt->mountdata = (char *) mnt_fs_get_fs_options(cxt->fs);
@@ -1042,7 +1103,7 @@ int mnt_context_next_mount(struct libmnt_context *cxt,
 	o = mnt_fs_get_user_options(*fs);
 	tgt = mnt_fs_get_target(*fs);
 
-	DBG(CXT, mnt_debug_h(cxt, "next-mount: trying %s", tgt));
+	DBG(CXT, ul_debugobj(cxt, "next-mount: trying %s", tgt));
 
 	/*  ignore swap */
 	if (mnt_fs_is_swaparea(*fs) ||
@@ -1062,7 +1123,7 @@ int mnt_context_next_mount(struct libmnt_context *cxt,
 					cxt->optstr_pattern))) {
 		if (ignored)
 			*ignored = 1;
-		DBG(CXT, mnt_debug_h(cxt, "next-mount: not-match "
+		DBG(CXT, ul_debugobj(cxt, "next-mount: not-match "
 				"[fstype: %s, t-pattern: %s, options: %s, O-pattern: %s]",
 				mnt_fs_get_fstype(*fs),
 				cxt->fstype_pattern,
@@ -1101,7 +1162,7 @@ int mnt_context_next_mount(struct libmnt_context *cxt,
 	}
 
 	if (mnt_context_is_child(cxt)) {
-		DBG(CXT, mnt_debug_h(cxt, "next-mount: child exit [rc=%d]", rc));
+		DBG(CXT, ul_debugobj(cxt, "next-mount: child exit [rc=%d]", rc));
 		DBG_FLUSH;
 		exit(rc);
 	}
